/* jshint undef: true, unused: true, jquery: true, browser: true */
/* global graph: true, alert, CatturaAPI, path, cytoscape, relativeRoot */

$(function () {
    //Variables
    var $graph = $("#cy");
    var cattura = new CatturaAPI(location.href + "/../" + relativeRoot);
    var columnClasses = ['1st_Row', '2nd_Row', '3rd_Row', '4th_Row', '5th_Row'];
    var nodesOrdered = graph.nodes.sort(compare);
    var currentNode_ID = 0;
    var plugins = [];
    
    /*
     * Discover what plugins are available
     */
    cattura.plugins.getPlugins().then(function (p) {
        for (var pluginID in p)
        {
            if (p[pluginID].title == pluginID)
            {
                p[pluginID].title = pluginID.substring(pluginID.lastIndexOf(".") + 1);
            }
        }
        plugins = p;
    });    
    
    //FUNCTIONS
    function condenseName (name) 
    {
        return name.substring(name.lastIndexOf(".") + 1);
    }
    
    function condenseTitle () 
    {
        var str = findNodeById(0).workingDir;
        var split = str.split("/");
        $("#title").empty().text(split[split.length - 1]);
    }
    
    function compare (a, b) 
    {
        if (a.nodeID < b.nodeID)
        {
            return -1;
        }
        
        if (a.nodeID > b.nodeID)
        {
            return 1;
        }
        return 0;
    }
    
    function switchMode (mode)
    {
        var $body = $("body");
        if (mode === "link")
        {
            $body.removeClass("mode-reconfigure").toggleClass("mode-link");
        }
        else if (mode === "reconfigure")
        {
            $body.removeClass("mode-link").toggleClass("mode-reconfigure");
        }
        else
        {
            $body.removeClass("mode-link mode-reconfigure");
        }
        
        if ($body.is(".mode-link")) $("#link-button").addClass("active");
        else $("#link-button").removeClass("active");

        if ($body.is(".mode-reconfigure")) $("#reconfigure-button").addClass("active");
        else $("#reconfigure-button").removeClass("active");
    }
    
    function makelevels (nodes) 
    {
        var x = 0, y = 0;
        var levels = [];
        
        // The node ids that need to be added to the visualizer.
        // Once a node id is added, it is removed from this list.
        // The list will be checked again to see if there are any
        // nodes that still need to be added (e.g. orphans, nodes without
        // parents, etc).
        var nodeIdsLeft = nodes.map(function (node) {
            return node.nodeID;
        });
        var currChild;
        
        //root setup
        levels.push([nodes[0].nodeID]);
        levels.push(nodes[0].connectedTo.slice());
        nodeIdsLeft.splice(nodeIdsLeft.indexOf(nodes[0].nodeID), 1);
        
        //rest
        for (x = 1; x < levels.length; x++) 
        {
            for (y = 0; y < levels[x].length; y++) 
            {
                nodeIdsLeft.splice(nodeIdsLeft.indexOf(levels[x][y]), 1);
                
                //get childeren of current
                currChild = findNodeById(levels[x][y]).connectedTo;
                if (levels[x + 1] === undefined)
                {
                    levels.push(currChild.slice());
                } 
                else
                {
                    levels[x + 1] = uniqueArray(levels[x + 1], currChild);
                }
            }
        }
        
        //one final level for orphaned nodes
        levels.push(nodes.filter(function (node) {
            if (node.connectedTo.length) return false;
            return !(nodes.some(function (parent) {
                return parent.connectedTo.indexOf(node.nodeID) > -1;
            }));
        }).map(function (node) {
            nodeIdsLeft.splice(nodeIdsLeft.indexOf(node.nodeID), 1);
            return node.nodeID;
        }));
        
        //nodes which have children but no parents
        //we need to put them in the graph
        /* jshint loopfunc:true */
        for (x = 1; x < levels.length; x++) 
        {
            for (y = 0; y < levels[x].length; y++) 
            {
                //We are looking for parents of nodes in the graph
                //which do not themselves have parents and were
                //not already added into the graph.
                var currentNode = findNodeById(levels[x][y]);
                var currNodeParents = nodes.filter(function (parent) {
                    return nodeIdsLeft.indexOf(parent.nodeID) > -1 && parent.connectedTo.indexOf(currentNode.nodeID) > -1;
                });
                
                for (var z = 0; z < currNodeParents.length; z++)
                {
                    if (!(nodes.some(function (parent) {
                        return parent.connectedTo.indexOf(currNodeParents[z].nodeID) > -1;
                    })))
                    {
                        var thisNodeId = currNodeParents[z].nodeID;
                        levels[x - 1].push(thisNodeId);
                        nodeIdsLeft.splice(nodeIdsLeft.indexOf(thisNodeId), 1);
                    }
                }
            }
        }
        /* jshint loopfunc:false*/
        
        return removeDupes(levels);
    }
    
    function makeNodesEdges () 
    {
        var columns = makelevels(nodesOrdered);
        var nodesEdges = [];
        var rootNode = findNodeById(0);
        
        //root set
        nodesEdges.push({
            group: "nodes",
            data: {id: $.parseJSON(rootNode.config).pluginID,
                node_id: 0,
                name: "root"},
            classes: 'root highlight',
            position: {x: 0, y: 0} // can alternatively specify position in rendered on-screen pixels 
        });

        var connects = rootNode.connectedTo;
        connects.forEach(function (value) {
            nodesEdges.push({
                group: 'edges',
                data: {
                    id: rootNode.nodeID + "_" + value,
                    source: $.parseJSON(rootNode.config).pluginID,
                    target: condenseName($.parseJSON(findNodeById(value).config).pluginID) + " " + value
                },
                classes: 'outEdge'
            });
        });

        //node sets
        var node, columnLeng, node_ID, shortName, startSetY, startSetX = 0;
        for (var x = 1; x < columns.length; x++) {
            columnLeng = columns[x].length;

            //Y-axis starting pos
            if (columnLeng % 2 === 0) 
            {
                startSetY = 50;
            } 
            else 
            {
                startSetY = 0;
            }
            
            //X-axis start pos
            startSetX += 250;

            for (var y = 0; y < columns[x].length; y++) 
            {
                node = findNodeById(columns[x][y]);
                shortName = condenseName($.parseJSON(node.config).pluginID);
                node_ID = node.nodeID;
                nodesEdges.push({
                    group: "nodes",
                    data: {
                        id: shortName + " " + node_ID,
                        node_id: node_ID,
                        name: shortName
                    },
                    classes: columnClasses[((x - 1) % columnClasses.length)],
                    position: {x: startSetX, y: startSetY} // can alternatively specify position in rendered on-screen pixels 
                });
                connects = node.connectedTo;
                
                /* jshint loopfunc:true */
                connects.forEach(function (value) {
                    nodesEdges.push({
                        group: 'edges',
                        data: {
                            id: node_ID + "_" + value,
                            source: shortName + " " + node_ID,
                            target: condenseName($.parseJSON(findNodeById(value).config).pluginID) + " " + value
                        }
                    });
                });
                /* jshint loopfunc:false */
                
                // next position
                if (startSetY > 0)
                {
                    startSetY *= -1;
                }
                else if (startSetY < 0)
                {
                    startSetY = (startSetY - 100) * -1;
                }
                else
                {
                    startSetY += 100;
                }
            }
        }
        return nodesEdges;
    }
    
    function removeDupes (stages)
    {
        var cleaned = stages;
        var val, index;
        for (var x = cleaned.length - 1; x > 0; x--) 
        {
            for (var y = 0; y <= cleaned[x].length; y++) 
            {
                val = cleaned[x][y];
                if (typeof val === "undefined" && cleaned[x].length === 0) 
                {
                    cleaned.splice(x, 1);
                    break;
                }
                
                for (var z = x - 1; z > 0; z--) 
                {
                    index = (cleaned[z]).indexOf(val);
                    if (index > -1) 
                    {
                        cleaned[z].splice(index, 1);
                        continue;
                    }
                }
            }
        }
        return cleaned;
    }
    
    function stateNStatus () 
    {
        var color = "#000000";
        if (currentNode_ID !== 0) {
            cattura.sessionOutput.getFile({
                root: path,
                nodeID: currentNode_ID,
                fileName: "state.json"
            }).then(function (data) {
                var stateText = data.state;
                
                if (data.stateMessage)
                {
                    stateText += " - " + data.stateMessage;
                }
                
                $("#state_status").empty().text(stateText);
                
                if (data.status === 'UNDETERMINED' || data.status === 'PARTIALLY_SUCCEEDED') 
                {
                    color = "#FF6F0A";
                } 
                else if (data.status === 'FAILED') 
                {
                    color = "#FF0000";
                } 
                else 
                {
                    color = "#00911A";
                }
                
                $("#status_status").empty().text(data.status).css({"color": color});
            });
        } 
        else 
        {
            $("#state_status").empty().text("NONE");
            $("#status_status").empty().text("NONE").css({"color": "#000000"});
        }
    }
    
    function syntaxHighlight (json) 
    {
        if (typeof json !== 'string') 
        {
            json = JSON.stringify(json, undefined, 2);
        }
        
        json = json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
        return json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
            var cls = 'number';
            if (/^"/.test(match)) 
            {
                if (/:$/.test(match)) 
                {
                    cls = 'key';
                } 
                else 
                {
                    cls = 'string';
                }
            } 
            else if (/true|false/.test(match))
            {
                cls = 'boolean';
            } 
            else if (/null/.test(match))
            {
                cls = 'null';
            }
            
            return '<span class="' + cls + '">' + match + '</span>';
        });
    }
    
    function uniqueArray (array1, array2) 
    {
        var unique = array1.concat(array2);
        for (var i = 0; i < unique.length; ++i) 
        {
            for (var j = i + 1; j < unique.length; ++j) 
            {
                if (unique[i] === unique[j]) 
                {
                    unique.splice(j--, 1);
                }
            }
        }
        return unique;
    }
    
    function findNodeById (id)
    {
        for (var i = 0; i < graph.nodes.length; i++)
        {
            if (graph.nodes[i].nodeID === id)
            {
                return graph.nodes[i];
            }
        }
    }

    function findPluginForNode (node)
    {
        var pluginID = JSON.parse(node.config).pluginID;
        var plugin = plugins[pluginID];
        return plugin;
    }
    
    function whatConnects (node_id) 
    {
        var what = [];
        var myNode = parseInt(node_id);
        for (var x = 0; x < nodesOrdered.length; x++) 
        {
            for (var y = 0; y < nodesOrdered[x].connectedTo.length; y++) 
            {
                if (myNode === nodesOrdered[x].connectedTo[y]) 
                {
                    what.push(x);
                }
            }
        }
        return what;
    }
    
    function reloadGraph ()
    {
        cattura.graphBuilder.getGraph({
            root: path
        }).then(function(g) {
            graph = g;
            nodesOrdered = graph.nodes.sort(compare);
            $graph.cytoscape('get').load(makeNodesEdges());
        });
    }
    
    /*
     * Graph-modifying functions
     */
    function linkNodes (fromNode, toNode)
    {
        var fromPlugin = findPluginForNode(findNodeById(fromNode));
        var toPlugin = findPluginForNode(findNodeById(toNode));
        var fromPluginName = fromPlugin ? fromPlugin.title : "Unknown Plugin";
        var toPluginName = toPlugin ? toPlugin.title : "Unknown Plugin";

        cattura.graphBuilder.linkNode({
            root: path,
            nodeID: fromNode,
            parentNodeIDs: "",
            childNodeIDs: toNode
        }).then(function(value) { 
            if (value)
            {
                if ($.inArray(toNode, value.connectedTo) > -1)
                {
                    cattura.graphBuilder.save({
                        root: path
                    }).then(function(value) {
                        if (value)
                        {
                            alert("Node " + fromNode + " (" + fromPluginName + ") was linked to node " + toNode + " (" + toPluginName + ") successfully.");
                            reloadGraph();
                        }
                    });
                }
                else
                {
                    alert("Node " + fromNode + " (" + fromPluginName + ") could not be linked to " + toNode + " (" + toPluginName + ") successfully.");
                }
            }
        });
    }
    
    // EVENTS
    $graph.cytoscape({
        style: cytoscape.stylesheet()
                .selector('node').css({'content': 'data(id)'})
                .selector('edge').css({'target-arrow-shape': 'triangle', 'width': 3, 'line-color': '#898898', 'target-arrow-color': '#898898'})
                .selector('.outEdge,.inEdge').css({"line-color": "#E714D7", 'target-arrow-color': '#E714D7'})
                .selector('.root').css({"background-color": "#000000"})
                .selector('.1st_Row').css({"background-color": "#269E06"})
                .selector('.2nd_Row').css({"background-color": "#132E7F"})
                .selector('.3rd_Row').css({"background-color": "#B4071B"})
                .selector('.4th_Row').css({"background-color": "#BB8307"})
                .selector('.5th_Row').css({"background-color": "#980655"})
                .selector('.highlight').css({"background-color": "#E714D7"}),
        elements: makeNodesEdges(),
        layout: {
            name: 'preset',
            fit: true,
            directed: true,
            roots: '#A',
            padding: 50
        },
        ready: function () {
            var cy = this;
            cy.on('tap', 'node', function () {
                $(".node_name,#modalTitle").text(this.id());
                var regex = /[0-9]*$/;
                var selectedNodeId = this.id().match(regex)[0];
                if (selectedNodeId === "") 
                {
                    selectedNodeId = 0;
                }
                else
                {
                    selectedNodeId = parseInt(selectedNodeId);
                }
                
                /*
                 * Link current node and selected node
                 */
                if ($("body").is(".mode-link"))
                {
                    linkNodes(currentNode_ID, selectedNodeId);
                    switchMode();
                    return;
                }
                
                currentNode_ID = selectedNodeId;                
                $('#json').trigger('click');
                stateNStatus();
                
                //highlight actual node
                cy.$(".highlight").removeClass('highlight');
                this.addClass('highlight');
                
                // highlight outs and ins edges
                var connectedTo = findNodeById(currentNode_ID).connectedTo;
                cy.$(".outEdge").removeClass('outEdge');
                for (var x = 0; x < connectedTo.length; x++) 
                {
                    cy.$("#" + currentNode_ID + "_" + connectedTo[x]).addClass('outEdge');
                }
                
                var connectedFrom = whatConnects(currentNode_ID);
                cy.$(".inEdge").removeClass('inEdge');
                for (x = 0; x < connectedFrom.length; x++) 
                {
                    cy.$("#" + connectedFrom[x] + "_" + currentNode_ID).addClass('inEdge');
                }

            }); //on tap
        }//onready
    }); //cy init
    
    //Buttons
    $("#json,#mJson").on('click', function () {
        switchMode();
        var obj = $.parseJSON(findNodeById(currentNode_ID).config);
        var json = JSON.stringify(obj, undefined, 5);
        var jsonHighlighted = "<pre>" + syntaxHighlight(json) + "</pre>";
        $("#deets_block, .modal_body_details").empty().html(jsonHighlighted);
        $("#node-reconfigure textarea").val(json);
    });
    
    $("#log,#mLog").on('click', function () {
        var $details = $("#deets_block,.modal_body_details").html("<h2>Loading logs...</h2>");
        switchMode();
        $.get(relativeRoot + "api/1/sessionOutput/" + path + "/nodes/" + currentNode_ID + "/processor.log", function (data) {
            if (data === "") 
            {
                $details.html("<h2>No Logs</h2>");
            } 
            else 
            {
                $("<span>").addClass("logs").text(data).appendTo($details.empty());
            }
        });
    });
    
    /*
     * Create Node Modal
     */    
    (function () {
        var $createNodeModal = $("#create-node");
        var $createNodePlugin = $createNodeModal.find("select");
        var $createNodeConfig = $createNodeModal.find("textarea").val("");
        
        $('#create-button').on("click", function () {
            switchMode();
            $createNodePlugin.find("optgroup").empty();
            for (var pluginID in plugins)
            {
                var plugin = plugins[pluginID];
                if (plugin.module === "media-processing-agent")
                {
                    $("<option>").attr("value", pluginID).text(plugin.title).appendTo($createNodePlugin.find("optgroup#create-node-plugin-" + plugin.type.toLowerCase()));                                        
                }
            }
            $createNodeConfig.val("{}");
            $createNodeModal.modal("show");
        });

        $createNodeModal.find(".modal-footer .btn-primary").click(function () {
            try
            {
                var pluginId = $createNodePlugin.val();
                var config = JSON.parse($createNodeConfig.val());
                cattura.graphBuilder.createNode({
                    root: path,
                    pluginID: pluginId,
                    config: config,
                    parentNodeIDs: ""
                }).then(function(value) { 
                    if (value)
                    {
                        cattura.graphBuilder.save({
                            root: path
                        }).then(function(value) {
                            if (value)
                            {
                                reloadGraph();
                                alert("The new node was created successfully");
                            }
                        });
                    }
                });
            }
            catch (e)
            {
                alert("Node configuration must be valid JSON.");
                return;
            }
        });
    })();
    
    /*
     * Remove Node Modal
     */
    (function () {
        var $removeNodeModal = $("#remove-node");

        $('#remove-button').on("click", function () {
            switchMode();
            $removeNodeModal.modal("show");
        });

        $removeNodeModal.find(".modal-footer .btn-primary").click(function () {
            cattura.graphBuilder.removeNode({
                root: path,
                nodeID: currentNode_ID
            }).then(function(value) { 
                if (value)
                {
                    cattura.graphBuilder.save({
                        root: path
                    }).then(function(value) {
                        if (value)
                        {
                            reloadGraph();
                            alert("The node was removed successfully");
                        }
                    });
                }
            });
        });
    })();
    
    /*
     * Link button
     */
    (function () {
        $('#link-button').click(function () {
            switchMode("link");
        });
    })();
    
    /*
     * Reconfigure button
     */
    (function () {
        $("#reconfigure-button").click(function () {
            switchMode("reconfigure");
        });
        
        $("#node-reconfigure button").click(function () {
            /*
             * Test configuration for JSONitude
             */
            try
            {
                var config = JSON.parse($("#node-reconfigure textarea").val());

                if (typeof config !== "object")
                {
                    alert("Node configuration must be an object.");
                    return;
                }

                /*
                 * Reconfigure the node using the graph builder API.
                 */
                var plugin = findPluginForNode(findNodeById(currentNode_ID));
                var pluginName = plugin ? plugin.title : "Unknown Plugin";
                cattura.graphBuilder.configureNode({
                    root: path,
                    nodeID: currentNode_ID,
                    config: config
                }).then(function(value) {
                    if (value)
                    {
                        cattura.graphBuilder.save({
                            root: path
                        }).then(function(value) {
                        if (value)
                        {
                            switchMode();
                            alert("Node " + currentNode_ID + " (" + pluginName + ") was reconfigured successfully");
                            reloadGraph();
                        }
                    });
                    }
                });
            }
            catch (e)
            {
                alert("Node configuration must be valid JSON.");
                return;
            }
        });
    })();    
    
    $("#output,#mOutput").on('click', function () {
        switchMode();
        cattura.sessionOutput.forNode({
            root: path,
            nodeID: currentNode_ID
        }).then(function (data) {
            var $listing = $('<div>');
            var output;
            var indent = 4;
            
            if (data.length === 0) 
            {
                $("#deets_block,.modal_body_details").empty().html("<h2>No Data</h2>");
            }
            else
            {
                for (var i = 0; i < data.length; i++)
                {
                    output = data[i];
                    $("<h3>").text(output.name).appendTo($listing);
                    var $div = $("<div>").append($("<b>Name:</b> " + output.name + "<br />"))
                        .append($("<b>Size:</b> " + output.size + " bytes<br />"))
                        .append($("<b>Type:</b> " + output.type + "<br />"))
                        .append($("<a href='" + relativeRoot + "api/1/sessionOutput/" + path + "/nodes/" + currentNode_ID + "/" + output.name + "'><b>Download</b></a><br />"));
                    if ('element' in output) {
                        $("<pre>").text(JSON.stringify(output.element, null, indent)).appendTo($div);
                    }
                    $("<hr>").appendTo($div);
                    $div.appendTo($listing);
                }
                $("#deets_block,.modal_body_details").empty().append($listing);
            }
        });
    });
    
    //Page Load Actions
    $("#json").trigger('click');
    condenseTitle();
    
    //Modal Actions
    $(window).resize(function () {
        $("#modalDetails").css({width: 'auto', height: 'auto'});
    });
    
    $(document).keyup(function (event) {
        if (event.which == 27)
        {
            switchMode();
        }
    });
});