/* jshint undef: true, unused: true */
/* global plugin, require, Packages */

var facade = require("facade");
var _ = require("lodash");
var enableEdits = true;
var on = require("on")(plugin);

var templates = {
    visualize: _(plugin.resource("templates", "visualize.tmpl").read().getVal()).template(),
    nograph: _(plugin.resource("templates", "nograph.tmpl").read().getVal()).template()
};

on({
    WebServerEvent: function (event)
    {
        var RequestResponseHelper = Packages.cattura.web.server.vos.RequestResponseHelper;
        var webServer = event.getWebServer();
        var environment = facade.retrieveProxy("EnvironmentProxy");
        var File = Packages.cattura.utils.io.wrappers.FileWrapper;
        var STORAGE = environment.getStorageDirPath().getCanonicalFile();

        webServer.add(function(request, response, control) {
            if (!request.uri().startsWith("/visualize/"))
            {
                control.nextHandler();
                return;
            }

            var helper = new RequestResponseHelper(request, response);

            if (helper.redirectStripDoubleSlashes())
            {
                return;
            }

            var path = request.uri().substring("/visualize/".length).replace("%20"," ");
            var relativeRoot = helper.getRelativeRoot();

            if (path.startsWith("js/") || path.startsWith("css/"))
            {
                var file = plugin.resource(path);
                if (file.isFile())
                {
                    response.header("Content-Type",file.type()).content(file.readBytes().getVal()).end();
                    return;
                }
                else
                {
                    control.nextHandler();
                    return;
                }
            }
            else if (path === "" || path === "/")
            {
                response.content(templates.nograph({
                    path: null
                })).end();
                return;
            }
            else
            {
                var graphTitle = path;
                var graphFile = new File(STORAGE.getAbsolutePath() + File.separator + path + File.separator + "graph.json");

                if (!graphFile.exists())
                {
                    response.content(templates.nograph({
                        path: path
                    })).end();
                    return;
                }

                var graph = graphFile.read().getOrRethrow();
                response.content(templates.visualize({
                    graphTitle: graphTitle,
                    request: request,
                    path: path,
                    graph: graph,
                    enableEdits: enableEdits,
                    relativeRoot: relativeRoot
                })).end();
            }
        });
    }
});
